/* I plan on trying to add this module to the curl dll and place
 * wrapperes in it that swig can interface to, so that we can
 * interface to Curl from other languages, etc. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <ctype.h>

#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#define _MPRINTF_REPLACE /* we want curl-functions instead of native ones */
#include <curl/mprintf.h>

#define CURLseparator	"--_curl_--"

/* This define make use of the "Curlseparator" as opposed to the
   MIMEseparator. We might add support for the latter one in the
   future, and that's why this is left in the source. */
#define CURL_SEPARATORS

/* This is now designed to have its own local setup.h */
#include "setup.h"

#if defined(WIN32)&&!defined(__CYGWIN32__)
#include <winsock.h>
#endif

#include "version.h"

#ifdef HAVE_IO_H /* typical win32 habit */
#include <io.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

/* The last #include file should be: */
#ifdef MALLOCDEBUG
/* this is low-level hard-hacking memory leak tracking shit */
#include "../lib/memdebug.h"
#endif

#define DEFAULT_MAXREDIRS  50L

#ifndef __cplusplus        /* (rabe) */
typedef char bool;
#endif                     /* (rabe) */

#define CURL_PROGRESS_STATS 0 /* default progress display */
#define CURL_PROGRESS_BAR   1

// Have some global vars in dll accessible through swig?
// These 4 will be used for progress callback.
unsigned int dltotal = 0;
unsigned int dlnow = 0;
unsigned int ultotal = 0;
unsigned int ulnow = 0;
/* Note: the following is a copy of later comments regarding use of global vars
 * to support the following curl options:
 * CURLOPT_PROXY		<host.domain.com> or <host.domain.com:port> (char *)
 * CURLOPT_PROXYPORT	(long)
 * CURLOPT_PROXYUSERPWD "username:passwd" (char *)
 * May want to set these and other config options from global vars
 * so we can just set them from java side without having to pass them
 * in function args.
 * The following options I would like to support this way (globals):
 * CURLOPT_VERBOSE ?		(boolean)
 * CURLOPT_FAILONERROR ?	(boolean)
 * CURLOPT_MUTE ?			(boolean)
 * CURLOPT_ERRORBUFFER		(char *buf[CURL_ERROR_SIZE])
 * CURLOPT_TIMEOUT			(long seconds)
 * CURLOPT_LOW_SPEED_LIMIT	(long bytes_per_second_threshold)
 * CURLOPT_LOW_SPEED_TIME	(long seconds_to_allow_under_lowspeed)
 * CURLOPT_INTERFACE		(char *if_name or hostname or ip_addr)
 */
char errbuf[CURL_ERROR_SIZE];
char proxy[255];
long proxyport;
char proxyuserpwd[255];
long timeout;
long lowspeed_limit;
long lowspeed_time;
char enet_interface[255];

/* Set this from script side to abort when progress callback returns */
int progret = 0;
/* Example of getting and setting these vars in java:
 * My_dltotal = curl.get_dltotal();
 * curl.set_progret(1); - Should make transfer abort
 * example.set_My_variable(example.get_My_variable() + n);
 */

int myprogress_callback(void *notused, size_t d_tot, size_t d_now, size_t u_tot, size_t u_now)
{
	dltotal = d_tot;
	dlnow = d_now;
	ultotal = u_tot;
	ulnow = u_now;
	return progret;
}

/***********************************************************************
 * Start with some silly functions to make win32-systems survive
 ***********************************************************************/
#if defined(WIN32) && !defined(__GNUC__) || defined(__MINGW32__)
static void win32_cleanup(void)
{
  WSACleanup();
}

static CURLcode win32_init(void)
{
  WORD wVersionRequested;  
  WSADATA wsaData; 
  int err; 
  wVersionRequested = MAKEWORD(1, 1); 
    
  err = WSAStartup(wVersionRequested, &wsaData); 
    
  if (err != 0) 
    /* Tell the user that we couldn't find a useable */ 
    /* winsock.dll.     */ 
    return CURLE_FAILED_INIT; 
    
  /* Confirm that the Windows Sockets DLL supports 1.1.*/ 
  /* Note that if the DLL supports versions greater */ 
  /* than 1.1 in addition to 1.1, it will still return */ 
  /* 1.1 in wVersion since that is the version we */ 
  /* requested. */ 
    
  if ( LOBYTE( wsaData.wVersion ) != 1 || 
       HIBYTE( wsaData.wVersion ) != 1 ) { 
    /* Tell the user that we couldn't find a useable */ 

    /* winsock.dll. */ 
    WSACleanup(); 
    return CURLE_FAILED_INIT; 
  }
  return CURLE_OK;
}
/* The Windows Sockets DLL is acceptable. Proceed. */ 
#else
static CURLcode win32_init(void) { return CURLE_OK; }
#define win32_cleanup()
#endif

/*
 * This is the main global constructor for the app. Call this before
 * _any_ libcurl usage. If this fails, *NO* libcurl functions may be
 * used, or havoc may be the result.
 */
// This function to be exported in DLL
CURLcode main_init(void)
{
  return win32_init();
}

/*
 * This is the main global destructor for the app. Call this after
 * _all_ libcurl usage is done.
 */
// This function to be exported in DLL
void main_free(void)
{
  win32_cleanup();
}

// This function to be exported in DLL
size_t read_data(void *ptr, size_t size, size_t nmemb, FILE *stream)
{
	size_t readed;
	readed = fread(ptr, size, nmemb, stream);
	return readed;
}

/* This function currently expects that main_init() should already be called.
 * it defaults to overwriting the file if it already exists.
 */
int putfile(char *filename, char *desturl, char *userpwd, long resume_offset)
{
	CURL *curl;
	int res;
	long http_code = 0;
	FILE *infd;
    int infilesize=-1; /* -1 means unknown */
    struct stat fileinfo;

	// Open file to send here?
	infd = (FILE *)fopen(filename,"rb");
    if (!infd || stat(filename, &fileinfo)) {
       return CURLE_READ_ERROR;
    }
    infilesize=fileinfo.st_size;

	//Init Curl
	curl = curl_easy_init();
	if(curl)
	{
		curl_easy_setopt(curl, CURLOPT_URL, desturl);
		curl_easy_setopt(curl, CURLOPT_PUT, TRUE);
		curl_easy_setopt(curl, CURLOPT_USERPWD, userpwd);
		curl_easy_setopt(curl, CURLOPT_INFILE, infd);
		curl_easy_setopt(curl, CURLOPT_READFUNCTION, read_data);
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, infilesize);
		curl_easy_setopt(curl, CURLOPT_PROGRESSFUNCTION, myprogress_callback);
		//curl_easy_setopt(curl, CURLOPT_PROGRESSDATA, NULL);
		curl_easy_setopt(curl, CURLOPT_MUTE, 1);
		curl_easy_setopt(curl, CURLOPT_VERBOSE, 0);
		if(resume_offset > 0)
		{
			// We add support for resuming
			curl_easy_setopt(curl, CURLOPT_RESUME_FROM, resume_offset);
		}
		/* Note: to support Proxy servers, we will must likely need
		 * to support the following curl options:
		 * CURLOPT_PROXY		<host.domain.com> or <host.domain.com:port> (char *)
		 * CURLOPT_PROXYPORT	(long)
		 * CURLOPT_PROXYUSERPWD "username:passwd" (char *)
		 * May want to set these and other config options from global vars
		 * so we can just set them from java side without having to pass them
		 * in function args.
		 * The following options I would like to support this way (globals):
		 * CURLOPT_VERBOSE ?		(boolean)
 		 * CURLOPT_FAILONERROR ?	(boolean)
		 * CURLOPT_MUTE ?			(boolean)
		 * CURLOPT_ERRORBUFFER		(char *buf[CURL_ERROR_SIZE])
		 * CURLOPT_TIMEOUT			(long seconds)
		 * CURLOPT_LOW_SPEED_LIMIT	(long bytes_per_second_threshold)
		 * CURLOPT_LOW_SPEED_TIME	(long seconds_to_allow_under_lowspeed)
		 * CURLOPT_INTERFACE		(char *if_name or hostname or ip_addr)
		 */
		
		res = curl_easy_perform(curl);

		res = curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &http_code);

		curl_easy_cleanup(curl);
	}
	fclose(infd);
	// Should reset global progress values here later.
	//return res;
	return (int)http_code;
}

// The length is set in c_length when we return from this call
// IF res returns SUCCESS, whatever that may be.
long remotefilelength(char *testurl, char *userpwd)
{
	CURL *curl;
	CURLcode res;
	long http_code = 0;
	double c_length = 0;
	FILE *hdfd;

	// Open file to send here?
	hdfd = (FILE *)fopen("curl_header.log","a");
    if (!hdfd) {
       return CURLE_READ_ERROR;
    }

	//Init Curl
	curl = curl_easy_init();
	if(curl)
	{
		curl_easy_setopt(curl, CURLOPT_URL, testurl);
		curl_easy_setopt(curl, CURLOPT_HEADER, 0);
		curl_easy_setopt(curl, CURLOPT_NOBODY, TRUE);
		curl_easy_setopt(curl, CURLOPT_USERPWD, userpwd);
		// Should add setopt calls that make curl quiet too
		curl_easy_setopt(curl, CURLOPT_MUTE, 1);
		curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1);
		curl_easy_setopt(curl, CURLOPT_VERBOSE, 0);
		//curl_easy_setopt(curl, CURLOPT_WRITEHEADER, hdfd);
		curl_easy_setopt(curl, CURLOPT_STDERR, hdfd);
		res = curl_easy_perform(curl);

		res = curl_easy_getinfo(curl, CURLINFO_HTTP_CODE, &http_code);
		// if http_code is 200, the file exists on remote, if it is 404, not found.
		res = curl_easy_getinfo(curl, CURLINFO_CONTENT_LENGTH_UPLOAD, &c_length);
		curl_easy_cleanup(curl);
	}
	fclose(hdfd);
	return (long)c_length;
}

